/*
 * Please do not edit this file.
 * It was generated using rpcgen.
 */

#include <memory.h> /* for memset */
#include "/build/libnvidia-container/src/libnvidia-container-1.19.1/src/nvc_rpc.h"
#pragma GCC diagnostic ignored "-Wmissing-prototypes"
#pragma GCC diagnostic ignored "-Wsign-conversion"
#pragma GCC diagnostic ignored "-Wunused-variable"

/* Default timeout can be changed using clnt_control() */
static struct timeval TIMEOUT = { 25, 0 };

enum clnt_stat 
driver_init_1(ptr_t arg1, driver_init_res *clnt_res,  CLIENT *clnt)
{
	return (clnt_call(clnt, DRIVER_INIT,
		(xdrproc_t) xdr_ptr_t, (caddr_t) &arg1,
		(xdrproc_t) xdr_driver_init_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
driver_shutdown_1(ptr_t arg1, driver_shutdown_res *clnt_res,  CLIENT *clnt)
{
	return (clnt_call(clnt, DRIVER_SHUTDOWN,
		(xdrproc_t) xdr_ptr_t, (caddr_t) &arg1,
		(xdrproc_t) xdr_driver_shutdown_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
driver_get_rm_version_1(ptr_t arg1, driver_get_rm_version_res *clnt_res,  CLIENT *clnt)
{
	return (clnt_call(clnt, DRIVER_GET_RM_VERSION,
		(xdrproc_t) xdr_ptr_t, (caddr_t) &arg1,
		(xdrproc_t) xdr_driver_get_rm_version_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
driver_get_cuda_version_1(ptr_t arg1, driver_get_cuda_version_res *clnt_res,  CLIENT *clnt)
{
	return (clnt_call(clnt, DRIVER_GET_CUDA_VERSION,
		(xdrproc_t) xdr_ptr_t, (caddr_t) &arg1,
		(xdrproc_t) xdr_driver_get_cuda_version_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
driver_get_device_count_1(ptr_t arg1, driver_get_device_count_res *clnt_res,  CLIENT *clnt)
{
	return (clnt_call(clnt, DRIVER_GET_DEVICE_COUNT,
		(xdrproc_t) xdr_ptr_t, (caddr_t) &arg1,
		(xdrproc_t) xdr_driver_get_device_count_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
driver_get_device_1(ptr_t arg1, u_int arg2, driver_get_device_res *clnt_res,  CLIENT *clnt)
{
	driver_get_device_1_argument arg;
	arg.arg1 = arg1;
	arg.arg2 = arg2;
	return (clnt_call (clnt, DRIVER_GET_DEVICE, (xdrproc_t) xdr_driver_get_device_1_argument, (caddr_t) &arg,
		(xdrproc_t) xdr_driver_get_device_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
driver_get_device_minor_1(ptr_t arg1, ptr_t arg2, driver_get_device_minor_res *clnt_res,  CLIENT *clnt)
{
	driver_get_device_minor_1_argument arg;
	arg.arg1 = arg1;
	arg.arg2 = arg2;
	return (clnt_call (clnt, DRIVER_GET_DEVICE_MINOR, (xdrproc_t) xdr_driver_get_device_minor_1_argument, (caddr_t) &arg,
		(xdrproc_t) xdr_driver_get_device_minor_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
driver_get_device_busid_1(ptr_t arg1, ptr_t arg2, driver_get_device_busid_res *clnt_res,  CLIENT *clnt)
{
	driver_get_device_busid_1_argument arg;
	arg.arg1 = arg1;
	arg.arg2 = arg2;
	return (clnt_call (clnt, DRIVER_GET_DEVICE_BUSID, (xdrproc_t) xdr_driver_get_device_busid_1_argument, (caddr_t) &arg,
		(xdrproc_t) xdr_driver_get_device_busid_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
driver_get_device_uuid_1(ptr_t arg1, ptr_t arg2, driver_get_device_uuid_res *clnt_res,  CLIENT *clnt)
{
	driver_get_device_uuid_1_argument arg;
	arg.arg1 = arg1;
	arg.arg2 = arg2;
	return (clnt_call (clnt, DRIVER_GET_DEVICE_UUID, (xdrproc_t) xdr_driver_get_device_uuid_1_argument, (caddr_t) &arg,
		(xdrproc_t) xdr_driver_get_device_uuid_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
driver_get_device_arch_1(ptr_t arg1, ptr_t arg2, driver_get_device_arch_res *clnt_res,  CLIENT *clnt)
{
	driver_get_device_arch_1_argument arg;
	arg.arg1 = arg1;
	arg.arg2 = arg2;
	return (clnt_call (clnt, DRIVER_GET_DEVICE_ARCH, (xdrproc_t) xdr_driver_get_device_arch_1_argument, (caddr_t) &arg,
		(xdrproc_t) xdr_driver_get_device_arch_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
driver_get_device_model_1(ptr_t arg1, ptr_t arg2, driver_get_device_model_res *clnt_res,  CLIENT *clnt)
{
	driver_get_device_model_1_argument arg;
	arg.arg1 = arg1;
	arg.arg2 = arg2;
	return (clnt_call (clnt, DRIVER_GET_DEVICE_MODEL, (xdrproc_t) xdr_driver_get_device_model_1_argument, (caddr_t) &arg,
		(xdrproc_t) xdr_driver_get_device_model_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
driver_get_device_brand_1(ptr_t arg1, ptr_t arg2, driver_get_device_brand_res *clnt_res,  CLIENT *clnt)
{
	driver_get_device_brand_1_argument arg;
	arg.arg1 = arg1;
	arg.arg2 = arg2;
	return (clnt_call (clnt, DRIVER_GET_DEVICE_BRAND, (xdrproc_t) xdr_driver_get_device_brand_1_argument, (caddr_t) &arg,
		(xdrproc_t) xdr_driver_get_device_brand_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
driver_get_device_mig_mode_1(ptr_t arg1, ptr_t arg2, driver_get_device_mig_mode_res *clnt_res,  CLIENT *clnt)
{
	driver_get_device_mig_mode_1_argument arg;
	arg.arg1 = arg1;
	arg.arg2 = arg2;
	return (clnt_call (clnt, DRIVER_GET_DEVICE_MIG_MODE, (xdrproc_t) xdr_driver_get_device_mig_mode_1_argument, (caddr_t) &arg,
		(xdrproc_t) xdr_driver_get_device_mig_mode_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
driver_get_device_max_mig_device_count_1(ptr_t arg1, ptr_t arg2, driver_get_device_max_mig_device_count_res *clnt_res,  CLIENT *clnt)
{
	driver_get_device_max_mig_device_count_1_argument arg;
	arg.arg1 = arg1;
	arg.arg2 = arg2;
	return (clnt_call (clnt, DRIVER_GET_DEVICE_MAX_MIG_DEVICE_COUNT, (xdrproc_t) xdr_driver_get_device_max_mig_device_count_1_argument, (caddr_t) &arg,
		(xdrproc_t) xdr_driver_get_device_max_mig_device_count_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
driver_get_device_mig_device_1(ptr_t arg1, ptr_t arg2, u_int arg3, driver_get_device_mig_device_res *clnt_res,  CLIENT *clnt)
{
	driver_get_device_mig_device_1_argument arg;
	arg.arg1 = arg1;
	arg.arg2 = arg2;
	arg.arg3 = arg3;
	return (clnt_call (clnt, DRIVER_GET_DEVICE_MIG_DEVICE, (xdrproc_t) xdr_driver_get_device_mig_device_1_argument, (caddr_t) &arg,
		(xdrproc_t) xdr_driver_get_device_mig_device_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
driver_get_device_gpu_instance_id_1(ptr_t arg1, ptr_t arg2, driver_get_device_gpu_instance_id_res *clnt_res,  CLIENT *clnt)
{
	driver_get_device_gpu_instance_id_1_argument arg;
	arg.arg1 = arg1;
	arg.arg2 = arg2;
	return (clnt_call (clnt, DRIVER_GET_DEVICE_GPU_INSTANCE_ID, (xdrproc_t) xdr_driver_get_device_gpu_instance_id_1_argument, (caddr_t) &arg,
		(xdrproc_t) xdr_driver_get_device_gpu_instance_id_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
driver_get_device_compute_instance_id_1(ptr_t arg1, ptr_t arg2, driver_get_device_compute_instance_id_res *clnt_res,  CLIENT *clnt)
{
	driver_get_device_compute_instance_id_1_argument arg;
	arg.arg1 = arg1;
	arg.arg2 = arg2;
	return (clnt_call (clnt, DRIVER_GET_DEVICE_COMPUTE_INSTANCE_ID, (xdrproc_t) xdr_driver_get_device_compute_instance_id_1_argument, (caddr_t) &arg,
		(xdrproc_t) xdr_driver_get_device_compute_instance_id_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
nvcgo_init_1(ptr_t arg1, nvcgo_init_res *clnt_res,  CLIENT *clnt)
{
	return (clnt_call(clnt, NVCGO_INIT,
		(xdrproc_t) xdr_ptr_t, (caddr_t) &arg1,
		(xdrproc_t) xdr_nvcgo_init_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
nvcgo_shutdown_1(ptr_t arg1, nvcgo_shutdown_res *clnt_res,  CLIENT *clnt)
{
	return (clnt_call(clnt, NVCGO_SHUTDOWN,
		(xdrproc_t) xdr_ptr_t, (caddr_t) &arg1,
		(xdrproc_t) xdr_nvcgo_shutdown_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
nvcgo_get_device_cgroup_version_1(ptr_t arg1, char *arg2, int arg3, nvcgo_get_device_cgroup_version_res *clnt_res,  CLIENT *clnt)
{
	nvcgo_get_device_cgroup_version_1_argument arg;
	arg.arg1 = arg1;
	arg.arg2 = arg2;
	arg.arg3 = arg3;
	return (clnt_call (clnt, NVCGO_GET_DEVICE_CGROUP_VERSION, (xdrproc_t) xdr_nvcgo_get_device_cgroup_version_1_argument, (caddr_t) &arg,
		(xdrproc_t) xdr_nvcgo_get_device_cgroup_version_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
nvcgo_find_device_cgroup_path_1(ptr_t arg1, int arg2, char *arg3, int arg4, int arg5, nvcgo_find_device_cgroup_path_res *clnt_res,  CLIENT *clnt)
{
	nvcgo_find_device_cgroup_path_1_argument arg;
	arg.arg1 = arg1;
	arg.arg2 = arg2;
	arg.arg3 = arg3;
	arg.arg4 = arg4;
	arg.arg5 = arg5;
	return (clnt_call (clnt, NVCGO_FIND_DEVICE_CGROUP_PATH, (xdrproc_t) xdr_nvcgo_find_device_cgroup_path_1_argument, (caddr_t) &arg,
		(xdrproc_t) xdr_nvcgo_find_device_cgroup_path_res, (caddr_t) clnt_res,
		TIMEOUT));
}

enum clnt_stat 
nvcgo_setup_device_cgroup_1(ptr_t arg1, int arg2, char *arg3, u_long arg4, nvcgo_setup_device_cgroup_res *clnt_res,  CLIENT *clnt)
{
	nvcgo_setup_device_cgroup_1_argument arg;
	arg.arg1 = arg1;
	arg.arg2 = arg2;
	arg.arg3 = arg3;
	arg.arg4 = arg4;
	return (clnt_call (clnt, NVCGO_SETUP_DEVICE_CGROUP, (xdrproc_t) xdr_nvcgo_setup_device_cgroup_1_argument, (caddr_t) &arg,
		(xdrproc_t) xdr_nvcgo_setup_device_cgroup_res, (caddr_t) clnt_res,
		TIMEOUT));
}
